#!/usr/bin/env python
###############################################################################
##
## MODULE      : postscript.py
## DESCRIPTION : Generate PostScript/PDF within GNU TeXmacs
## COPYRIGHT   : (C) 2019  Darcy Shen
##
## This software falls under the GNU general public license version 3 or later.
## It comes WITHOUT ANY WARRANTY WHATSOEVER. For details, see the file LICENSE
## in the root directory or <http://www.gnu.org/licenses/gpl-3.0.html>.

import os
import platform
from io import BytesIO
from io import open
from .protocol import *


class PSOutDummy:
    """ Dummy class for use with ps_out.

    We return an instance of this class to avoid output after 
    evaluation in the TeXmacs plugin of ps_out."""

    def __init__(self, data):
        self.content = data

    def __str__(self):
        """Return an empty string for compose_output()"""
        return ''

    def __repr__(self):
        return 'PSOutDummy'

class FileOutDummy:
    """ Python object to feed a file back to TeXmacs. """

    def __init__(self, data):
        self.content = data

    def __str__(self):
        """Return an empty string for compose_output()"""
        return ''

    def __repr__(self):
        return 'FileOutDummy'


def ps_out (out):
    """Outputs PostScript within TeXmacs.

    According the the type of the argument the following
    scenarios can take place:    

    If the argument is an instance of matplotlib.pyplot.Figure
    then its method savefig() will be used to produce an EPS
    figure. Note that you need to be using a backend which
    supports this format.

    If the argument is a string and has more than one line, it
    will be processed as raw Postscript data.

    If the argument is a string with no line breaks, it is assumed
    to contain the filename of a Postscript file which will be
    read (if the file  has no extension, the defaults .eps and .ps
    will be tried in this order).

    If the argument is a file or any other object which provides
    a 'read'  method, data will be obtained by calling such
    method.

    Implemented from suggestion by Alvaro Tejero Cantero.
    Implementation partially based on information provided
    by Mark Arrasmith.
    """
    if 'savefig' in dir(out):
        str_out = BytesIO()
        out.savefig(str_out, format='eps')
        data = str_out.getvalue()
        str_out.close()
    elif isinstance(out, str):
        if out.find('\n') > 0:
            data = out
        else:
            ext_list = ['', '.eps', '.ps']
        for ext in ext_list:
            if os.path.exists(out+ext):
                fd = open(out+ext, 'rb')
                data = fd.read()
                fd.close()
                break
            else:
                raise IOError('File "%s%s" not found.' % (out, str(ext_list)))
    elif 'read' in dir(out):
        data = out.read()

    return PSOutDummy(texmacs_escape(data).decode())

pdf_out_tmp_file = "pdf_out_" + str(os.getpid()) + ".pdf"
if (platform.system() == "Windows"):
    pdf_out_tmp_file = os.getenv("TEXMACS_HOME_PATH") + "\\system\\tmp\\" + pdf_out_tmp_file
else:
    pdf_out_tmp_file = os.getenv("TEXMACS_HOME_PATH") + "/system/tmp/" +  pdf_out_tmp_file


def pdf_out (out):
    """Outputs PDF within TeXmacs.

    According the the type of the argument the following
    scenarios can take place:

    If the argument is an instance of matplotlib.pyplot.Figure
    then its method savefig() will be used to produce a PDF
    figure.
    Note that you need to be using a backend which supports this format.

    If the argument is a string with no line breaks, it is assumed
    to contain the filename of a PDF file which will be
    passed to TeXmacs (if the file  has no extension, the default
    .pdf will be tried).

    If the argument is a file or any other object which provides
    a 'read'  method, data will be obtained by calling such
    method.
    
    In any case a (maybe temporary) PDF filename is passed to TeXmacs.
    Adapted from ps_out().
    """

    if 'savefig' in dir(out):
        out.savefig(pdf_out_tmp_file, format='pdf')
        name = pdf_out_tmp_file
    elif isinstance(out, str):
        if out.find('\n') > 0:
            name = None
        else:
            ext_list = ['', '.pdf']
        for ext in ext_list:
            if os.path.exists(out+ext):
                name = out+ext
                break
            else:
                raise IOError('File "%s%s" not found.' % (out, str(ext_list)))
    elif 'read' in dir(out):
        data = out.read()
        fd = open(pdf_out_tmp_file, 'wb')
        fd.write(data)
        fd.close()
        name = pdf_out_tmp_file

    return FileOutDummy(name)



#!/usr/bin/env python
###############################################################################
##
## MODULE      : xypic.py
## DESCRIPTION : XYpic plotting support
## COPYRIGHT   : (C) 2019  Darcy Shen
##               (C) 2004 Nicolas Ratier (nicolas DOT ratier AT lpmo DOT edu)
##               (C) XYpic latex package Kristoffer H. Rose
##
## This software falls under the GNU general public license version 3 or later.
## It comes WITHOUT ANY WARRANTY WHATSOEVER. For details, see the file LICENSE
## in the root directory or <http://www.gnu.org/licenses/gpl-3.0.html>.

import os
from subprocess import Popen, PIPE, STDOUT
from .latex import LaTeX
from ..protocol import *


class XYpic(LaTeX):
    def __init__(self, name = "xypic"):
        super(XYpic, self).__init__()
        self.name = name

        self.pre_code = """
\\documentclass{standalone}
\\usepackage[all]{xy}
\\usepackage{amsmath}
\\usepackage{amssymb}
\\begin{document}
"""
        self.post_code = "\end{document}"
        self.message = "TeXmacs interface to XYpic (high level 2-dimensional graphics)"

    def evaluate(self, code):
        if not (code.lstrip().startswith("\\documentclass")):
            code = self.pre_code + "\n" + code + "\n" + self.post_code

        super(XYpic, self).evaluate(code)




#!/usr/bin/env python
###############################################################################
##
## MODULE      : plantuml.py
## DESCRIPTION : PlantUML plotting support
## COPYRIGHT   : (C) 2019  Darcy Shen
##
## This software falls under the GNU general public license version 3 or later.
## It comes WITHOUT ANY WARRANTY WHATSOEVER. For details, see the file LICENSE
## in the root directory or <http://www.gnu.org/licenses/gpl-3.0.html>.

import os
from subprocess import Popen, PIPE, STDOUT
from .graph import Graph
from ..protocol import *

class PlantUML(Graph):
    def __init__(self, name = "plantuml"):
        super(PlantUML, self).__init__()
        self.name = name

    def greet(self):
        if len(self.message) == 0:
            try:
                p = Popen([self.name, "-version"], stdout=PIPE)
                ret, err = p.communicate()
                if (p.returncode == 0):
                    self.message = ret.decode()
            except OSError:
                pass
        super(PlantUML, self).greet()

    def evaluate(self, code):
        code_path = self.get_tmp_dir() + self.name + ".puml" 
        with open(code_path, 'w') as code_file:
            code_file.write(code)

        # Execute `plantuml -teps /path/to/plantuml.puml`
        cmd = [self.name, "-teps", code_path]
        p = Popen(cmd, stderr=PIPE)
        out, err = p.communicate()
        if (p.returncode == 0):
            flush_file (self.get_eps())
        else:
            flush_verbatim (err.decode())
        



#!/usr/bin/env python
###############################################################################
##
## MODULE      : feynmf.py
## DESCRIPTION : FeynMF graphics support
## COPYRIGHT   : (C) 2019  Darcy Shen and Massimiliano Gubinelli
##
## This software falls under the GNU general public license version 3 or later.
## It comes WITHOUT ANY WARRANTY WHATSOEVER. For details, see the file LICENSE
## in the root directory or <http://www.gnu.org/licenses/gpl-3.0.html>.

## Inspired by the tm_feynmf pluging written by M. R. Wegewijs 

import os
from subprocess import Popen, PIPE, STDOUT
from .graph import Graph
from ..compat import which
from ..protocol import *

# we cannot derive from the LaTeX object since we need more complex processing 
# than latex->dvips

class FeynMF(Graph):
    def __init__(self, name = "feynmf"):
        super(FeynMF, self).__init__()
        self.name = name
        self.message = "TeXmacs interface to the FeynMF picture generator"

    def available(self):
        return all(list(map(lambda x: which(x) is not None, ["latex", "dvips", "mpost", "gs"])))

    def evaluate(self, code):

        if not (code.lstrip().startswith("\\documentclass")):
            code = ("\\documentclass{article}\n" ## feynmf does not work with the standalone class
                    "\\usepackage{feynmp}\n"
                    "\\pagestyle{empty}\n"    
                    "\\unitlength=20mm\n"
                    "\\begin{document}\n" 
                    "\\begin{fmffile}{fmftempl}\n"   
                    # add syles    
                    "\\begin{fmfgraph*}"        
                    ) + code + "\n\\end{fmfgraph*}\n\\end{fmffile}\n\\end{document}\n"   

        code_path = self.get_tmp_dir() + self.name + ".tex"
        dvi_path = self.get_tmp_dir() + self.name + ".dvi"
        ps2_path = self.get_tmp_dir() + self.name + ".ps"

        with open(code_path, 'w') as code_file:
            code_file.write(code)

        cmd0 = ["latex", "--interaction=nonstopmode", "-halt-on-error", code_path ] 
        cmd1 = ["dvips", "-q", dvi_path, "-o", ps2_path ]
        cmd2 = ["mpost", "--interaction=nonstopmode", "-halt-on-error", "fmftempl.mp" ]
#        cmd2 = ["mf", "--interaction=nonstopmode", "-halt-on-error", "\mode:=localfont; input fmftempl" ]
        cmd3 = ["gs", "-o", self.get_eps_path(), "-sDEVICE=eps2write" , "-dEPSCrop" , ps2_path ]

        # flush_err (code) # to debug
        phase = "LaTeX (pass 1/2)"    
        #flush_err ("Running " + phase)
        os.chdir(self.get_tmp_dir())
        p = Popen(cmd0, stdout=PIPE, stderr=STDOUT)
        out, err = p.communicate()
        if (0 == 0): # (p.returncode == 0):
            phase = "MetaPost"    
            #flush_err ("Running " + phase)
            p = Popen(cmd2, stdout=PIPE, stderr=STDOUT)
            out, err = p.communicate()
            if (p.returncode == 0):
                phase = "LaTeX (pass 2/2)"    
                #flush_err ("Running " + phase)
                p = Popen(cmd0, stdout=PIPE, stderr=STDOUT)
                out, err = p.communicate()
                if (p.returncode == 0):
                    phase = "dvips"    
                    #flush_err ("Running " + phase)
                    p = Popen(cmd1, stdout=PIPE, stderr=STDOUT)
                    out, err = p.communicate()
                    if (p.returncode == 0):
                        phase = "Distilling EPS"    
                        #flush_err ("Running " + phase)
                        p = Popen(cmd3, stdout=PIPE, stderr=STDOUT)
                        out, err = p.communicate()
                        if (p.returncode == 0):
                            phase = "Finished"    
                            #flush_err ("Running " + phase)
                            flush_file (self.get_eps())
                            return
        flush_err (phase + " error!\n" + out.decode())
        flush_verbatim("")   # otherwise TeXmacs keeps waiting output



#!/usr/bin/env python
###############################################################################
##
## MODULE      : gnuplot.py
## DESCRIPTION : Gnuplot plotting support
## COPYRIGHT   : (C) 2019  Darcy Shen
##
## This software falls under the GNU general public license version 3 or later.
## It comes WITHOUT ANY WARRANTY WHATSOEVER. For details, see the file LICENSE
## in the root directory or <http://www.gnu.org/licenses/gpl-3.0.html>.

import os
from subprocess import Popen, PIPE, STDOUT
from .graph     import Graph
from ..protocol import *
from ..compat   import *

class Gnuplot(Graph):
    def __init__(self, name = "gnuplot"):
        super(Gnuplot, self).__init__()
        self.name = name
        self.default_output = "svg"
        self.default_width = "0.8par"

    def greet(self):
        if len(self.message) == 0:
            try:
                p = Popen([self.name, "-V"], stdout=PIPE)
                ret, err = p.communicate()
                # WARN: The Version Info is in stderr
                if (p.returncode == 0):
                    self.message = ret.decode()
            except OSError:
                pass
        super(Gnuplot, self).greet()

    def evaluate(self, code):
        if self.output == "eps":
            image = self.get_eps()
            self.pre_code = """
reset
set terminal postscript eps enhanced
set output 
set output '%s'
set size 1,1
set autoscale
""" % (self.get_eps_path())
        elif self.output == "png":
            image = self.get_png()
            self.pre_code = """
reset
set terminal pngcairo enhanced
set output 
set output '%s'
set size 1,1
""" % (self.get_png_path())
        elif self.output == "svg":
            image = self.get_svg()
            self.pre_code = """
reset
set terminal svg enhanced
set output 
set output '%s'
set size 1,1
set autoscale
""" % (self.get_svg_path())
        else:
            flush_verbatim("Unsupported output type: " + self.output)
            return

        code_path = self.get_tmp_dir() + self.name + ".txt"
        with open(code_path, 'w') as code_file:
            code_file.write(self.pre_code)
            code_file.write(code)

        cmd = [self.name, "-c", code_path]
        p = Popen(cmd, stderr=PIPE)
        out, err = p.communicate()
        if (p.returncode == 0):
          flush_file (image)
        else:
          flush_verbatim (err.decode())



#!/usr/bin/env python
###############################################################################
##
## MODULE      : diagrams.py
## DESCRIPTION : Diagrams plotting support
## COPYRIGHT   : (C) 2019  Darcy Shen
##
## This software falls under the GNU general public license version 3 or later.
## It comes WITHOUT ANY WARRANTY WHATSOEVER. For details, see the file LICENSE
## in the root directory or <http://www.gnu.org/licenses/gpl-3.0.html>.

import os
from subprocess import Popen, PIPE, STDOUT
from .graph     import Graph
from ..protocol import *
from ..compat   import *

class Diagrams(Graph):
    def __init__(self, name):
        super(Diagrams, self).__init__()
        self.name = name
        self.width = 600

    def greet(self):
        if len(self.message) == 0:
            try:
                p = Popen(["diagrams", "--version"], stdout=PIPE)
                ret, err = p.communicate()
                if (p.returncode == 0):
                    self.message = ret.decode()
            except OSError:
                pass
        super(Diagrams, self).greet()

    def available(self):
        return which("diagrams") is not None

    def evaluate(self, code):
        code_path = self.get_tmp_dir() + self.name + ".txt"
        with open(code_path, 'w') as code_file:
            code_file.write(code)

        cmd = ["diagrams", self.name, code_path, self.get_svg_path()]
        p = Popen(cmd, stderr=PIPE)
        out, err = p.communicate()
        if (p.returncode == 0):
          flush_file (self.get_svg())
        else:
          flush_verbatim (err.decode())



#!/usr/bin/env python
###############################################################################
##
## MODULE      : complete.py
## DESCRIPTION : Complete implemented in Python
## COPYRIGHT   : (C) 2019  Darcy Shen
##
## This software falls under the GNU general public license version 3 or later.
## It comes WITHOUT ANY WARRANTY WHATSOEVER. For details, see the file LICENSE
## in the root directory or <http://www.gnu.org/licenses/gpl-3.0.html>.

import os
import sys
from os.path import exists
tmpy_home_path = os.environ.get("TEXMACS_HOME_PATH") + "/plugins/tmpy"
if (exists (tmpy_home_path)):
    sys.path.append(os.environ.get("TEXMACS_HOME_PATH") + "/plugins/")
else:
    sys.path.append(os.environ.get("TEXMACS_PATH") + "/plugins/")

from tmpy.protocol        import *
from tmpy.compat          import *

flush_verbatim ("We know how to complete 'h'")

while True:
    line = tm_input()
    if not line:
        pass
    if line[0] == DATA_COMMAND:
        flush_scheme ("""(tuple "h" "ello" "i there" "ola" "opsakee")""")
    else:
        flush_verbatim ("You typed " + line)



#!/usr/bin/env python
###############################################################################
##
## MODULE      : prompt.py
## DESCRIPTION : Prompt implemented in Python
## COPYRIGHT   : (C) 2019  Darcy Shen
##
## This software falls under the GNU general public license version 3 or later.
## It comes WITHOUT ANY WARRANTY WHATSOEVER. For details, see the file LICENSE
## in the root directory or <http://www.gnu.org/licenses/gpl-3.0.html>.

import os
import sys
from os.path import exists
tmpy_home_path = os.environ.get("TEXMACS_HOME_PATH") + "/plugins/tmpy"
if (exists (tmpy_home_path)):
    sys.path.append(os.environ.get("TEXMACS_HOME_PATH") + "/plugins/")
else:
    sys.path.append(os.environ.get("TEXMACS_PATH") + "/plugins/")

from tmpy.protocol        import *
from tmpy.compat          import *

flush_verbatim ("A LaTeX -> TeXmacs converter")

counter = 0
while True:
    counter= counter + 1
    flush_prompt ("Input " + str(counter) + "] ")
    line = tm_input ()
    if not line:
        pass
    else:
        flush_latex (line)



#!/usr/bin/env python
###############################################################################
##
## MODULE      : tm_plantuml.py
## DESCRIPTION : Launcher for the PlantUML plugin
## COPYRIGHT   : (C) 2020  Darcy Shen
##
## This software falls under the GNU general public license version 3 or later.
## It comes WITHOUT ANY WARRANTY WHATSOEVER. For details, see the file LICENSE
## in the root directory or <http://www.gnu.org/licenses/gpl-3.0.html>.

import os
import sys
from os.path import exists
tmpy_home_path = os.environ.get("TEXMACS_HOME_PATH") + "/plugins/tmpy"
if (exists (tmpy_home_path)):
    sys.path.append(os.environ.get("TEXMACS_HOME_PATH") + "/plugins/")
else:
    sys.path.append(os.environ.get("TEXMACS_PATH") + "/plugins/")

from tmpy.graph.plantuml  import PlantUML
from tmpy.capture         import CaptureStdout
from tmpy.protocol        import *
from tmpy.compat          import *


if (exists (tmpy_home_path)):
    flush_verbatim ("WARNING: You are under develop mode using " + tmpy_home_path)
    flush_newline (2)

my_globals   = {}

if py_ver == 3:
    text = 'import builtins as __builtins__'
else:
    text = 'import __builtin__ as __builtins__'
CaptureStdout.capture (text, my_globals, "tm_plantuml")

current = PlantUML()
current.greet()
current.main_loop()



#!/usr/bin/env python
###############################################################################
##
## MODULE      : tm_graph.py
## DESCRIPTION : Launcher for the Graph plugin
## COPYRIGHT   : (C) 2004       Ero Carrera, ero@dkbza.org
##               (C) 2012       Adrian Soto
##               (C) 2014       Miguel de Benito Delgado, mdbenito@texmacs.org
##               (C) 2018-2019  Darcy Shen
##
## This software falls under the GNU general public license version 3 or later.
## It comes WITHOUT ANY WARRANTY WHATSOEVER. For details, see the file LICENSE
## in the root directory or <http://www.gnu.org/licenses/gpl-3.0.html>.

import os
import sys
from os.path import exists
tmpy_home_path = os.environ.get("TEXMACS_HOME_PATH") + "/plugins/tmpy"
if (exists (tmpy_home_path)):
    sys.path.append(os.environ.get("TEXMACS_HOME_PATH") + "/plugins/")
else:
    sys.path.append(os.environ.get("TEXMACS_PATH") + "/plugins/")

import re
import string
import time
import csv   # Used to parse scheme forms
from inspect   import ismodule, getsource, getsourcefile
from types     import CodeType
from io        import open
from io        import StringIO
from tmpy.protocol        import *
from tmpy.capture         import CaptureStdout
from tmpy.graph.asymptote import Asymptote
from tmpy.graph.dratex    import DraTeX
from tmpy.graph.graphviz  import Graphviz
from tmpy.graph.mermaid   import Mermaid
from tmpy.graph.plantuml  import PlantUML
from tmpy.graph.xypic     import XYpic
from tmpy.graph.tikz      import TikZ
from tmpy.graph.feynmf    import FeynMF
from tmpy.graph.latex     import LaTeX
from tmpy.graph.pdflatex  import PDFLaTeX
from tmpy.graph.gnuplot   import Gnuplot
from tmpy.graph.diagrams  import Diagrams
from tmpy.compat          import *

#import logging as log
#log.basicConfig(filename='/tmp/tm_python.log',level=log.DEBUG)

init_time = time.time()

__version__  = '1.0'
__author__   = 'Darcy Shen'

my_globals   = {}

def compose_output(data):
    """Do some parsing on the output according to its type.
    
    Non printable characters in unicode strings are escaped
    and objects of type None are not printed (so that procedure calls,
    as opposed to function calls, don't produce any output)."""

    if py_ver == 3: cl = str
    else:           cl = unicode
    if isinstance(data, cl):
        data2 = r''
        for c in data:
            if c not in string.printable:
                data2 += '\\x%x' % ord(c)
            else:
                data2 += c
        data = data2
    if data is None:
        data = ''
    return 'verbatim:%s' % str(data).strip()

def as_scm_string (text):
    return '"%s"' % text.replace('\\', '\\\\').replace('"', '\\"')

###############################################################################
## Session start
###############################################################################

if py_ver == 3:
    text = 'import builtins as __builtins__'
else:
    text = 'import __builtin__ as __builtins__'
CaptureStdout.capture (text, my_globals, "tm_graph")

# Reopen stdout unbufferd (flush after each stdout.write() and print)
if py_ver == 3:
    sys.stdout = os.fdopen (sys.stdout.fileno(), 'w')
else:
    sys.stdout = os.fdopen (sys.stdout.fileno(), 'w', 0)

grapvizs = list(map(lambda x: Graphviz(x), ["dot", "neato", "twopi", "circo", "fdp", "sfdp", "patchwork", "osage"]))
diagrams = list(map(lambda x: Diagrams(x), ["flowchart", "sequence", "railroad"]))
others = [Asymptote(), PlantUML(), Mermaid(), XYpic(), TikZ(),
          FeynMF(), LaTeX(), PDFLaTeX(), Gnuplot(), DraTeX()]
graphs = list(filter(lambda x: x.available(), grapvizs + diagrams + others))
graph_names = list(map(lambda x: x.name, graphs))

if len(graphs) == 0:
    flush_newline ()
    flush_verbatim ("Sorry, please check your installation of any of:")
    flush_verbatim ("Graphviz/Diagrams/LaTeX/Gnuplot/Asymptote/PlantUML/Mermaid")
    flush_prompt ("dead] ")
    exit(0)

current = graphs[0]
warmup_time_ms = int((time.time() - init_time) * 1000)

if (exists (tmpy_home_path)):
    flush_verbatim ("WARNING: You are under develop mode using " + tmpy_home_path)
    flush_newline (2)
flush_verbatim ("Generate graphs with your favorite tools in GNU TeXmacs\n")
flush_verbatim ("Created by Darcy Shen, Implemented in Python, " + current.name + "[" + str(warmup_time_ms) + "ms] by default\n")
flush_verbatim ("Welcome to star and fork it at https://github.com/texmacs/plugins\n")
flush_prompt (current.name + "] ")

def unigraph(text):
    magic_lines = text.split("\n")
    magic_line = magic_lines[0]
    command = magic_line.split(" ")[0].strip("%")

    if command in graph_names:
        graph = graphs[graph_names.index(command)]
        graph.eval(text)
    else:
        flush_verbatim ("No such Graphs backend: " + command)


# Main session loop.
while True:
    line = tm_input()
    if not line:
        continue
    if line[0] == DATA_COMMAND:
        # TODO: Handle completions
        continue
    else:
        lines = [line]
        while line != "<EOF>":
            line = tm_input()
            if line == '': 
                continue
            lines.append(line)

        text='\n'.join(lines[:-1])

        if text.startswith("%"):
            unigraph(text)
        elif text == "help":
            flush_verbatim ("[help, " + ", ".join(str(x) for x in graph_names) + "]\n")
            current.greet()
        elif text in graph_names:
            current = graphs[graph_names.index(text)]
            current.greet()
        else:
            current.eval(text)



#!/usr/bin/env python
###############################################################################
##
## MODULE      : asymptote.py
## DESCRIPTION : Asymptote plotting support
## COPYRIGHT   : (C) 2019  Darcy Shen
##
## This software falls under the GNU general public license version 3 or later.
## It comes WITHOUT ANY WARRANTY WHATSOEVER. For details, see the file LICENSE
## in the root directory or <http://www.gnu.org/licenses/gpl-3.0.html>.

import os
from subprocess import Popen, PIPE, STDOUT
from .graph import Graph
from ..protocol import *
from ..compat import *

class Asymptote(Graph):
    def __init__(self, name = "asy"):
        super(Asymptote, self).__init__()
        self.name = name

    def greet(self):
        if len(self.message) == 0:
            try:
                p = Popen([self.name, "-version"], stderr=PIPE)
                ret, err = p.communicate()
                # WARN: The Version Info is in stderr
                if (p.returncode == 0):
                    self.message = err.decode()
            except OSError:
                pass
        super(Asymptote, self).greet()

    def evaluate(self, code):
        code_path = self.get_tmp_dir() + self.name + ".asy"
        with open(code_path, 'w') as code_file:
            code_file.write(code)

        cmd = [self.name, "-quiet", "-feps", "-o", self.get_eps_path(), code_path]
        p = Popen(cmd, stderr=PIPE)
        out, err = p.communicate()
        if (p.returncode == 0):
          flush_file (self.get_eps())
        else:
          err_msg = err.decode()
          flush_verbatim (err_msg)
          # extra notification of missing semicolon
          if "unexpected end of input" in err_msg or "does not exist" in err_msg:
            flush_verbatim ("\nNote: This error might be related to your input, such as a missing semicolon.")

    def main_loop(self):
        # Main session loop.
        while True:
            line = tm_input()
            if not line:
                continue
            if line[0] == DATA_COMMAND:
                # TODO: Handle completions
                continue
            else:
                lines = []
                for x in line.split('~'):
                    lines.append(x)
                while line != "<EOF>":
                    line = tm_input()
                    for x in line.split('~'):
                        lines.append(x)
                text='\n'.join(lines[:-1])
                self.eval(text)

